/*----------------------------------------------------------------------
  File    : clc.c
  Contents: probabilistic and fuzzy cluster comparison
  Author  : Christian Borgelt
  History : 2006.05.28 file created from file cle.c
            2006.08.14 normalization of coincidence matrix entries added
            2007.02.14 adapted to changed module tabread
            2008.04.08 adapted to modified cluster set interface
            2008.09.21 adapted to use of mat_match (Hungarian method)
            2013.08.16 adapted to definitions ATTID, VALID, DIMID etc.
            2013.08.29 adapted to new function as_target()
            2014.10.24 changed from LGPL license to MIT license
            2018.08.18 extended functions (attset/table) removed
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <assert.h>
#ifndef MAT_RDWR
#define MAT_RDWR
#endif
#include "matrix.h"
#ifndef AS_READ
#define AS_READ
#endif
#ifndef AS_WRITE
#define AS_WRITE
#endif
#ifndef AS_PARSE
#define AS_PARSE
#endif
#include "attset.h"
#include "attmap.h"
#ifndef CLS_PARSE
#define CLS_PARSE
#endif
#include "cluster.h"
#include "error.h"
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define PRGNAME     "clc"
#define DESCRIPTION "probabilistic and fuzzy cluster comparison"
#define VERSION     "version 3.3 (2022.04.13)         " \
                    "(c) 2006-2022   Christian Borgelt"

#define BLKSIZE     256         /* block size for column vectors */

/* --- conjunction functions --- */
#define PM_MIN        0         /* minimum */
#define PM_MINNP      1         /* nil-potent minimum */
#define PM_PROD       2         /* product */
#define PM_LUKA       3         /* Lukasiewicz */
#define PM_CFCNT      4         /* number of functions */

/* --- evaluation measures --- */
#define PM_DIFF       0         /* sum of squared differences */
#define PM_ACC        1         /* cross-classification accuracy */
#define PM_F1         2         /* F1 measure */
#define PM_RAND       3         /* Rand statistic */
#define PM_JACCARD    4         /* Jaccard coefficient */
#define PM_FOLKES     5         /* Folkes-Mallows index */
#define PM_HUBERT     6         /* Hubert index */

/* --- error codes --- */
/* error codes 0 to -5 defined in attset.h */
#define E_OPTION     (-6)       /* unknown option */
#define E_OPTARG     (-7)       /* missing option argument */
#define E_ARGCNT     (-8)       /* wrong number of arguments */
#define E_PARSE      (-9)       /* parse errors on input file */
#define E_ATTCNT    (-10)       /* no usable attributes found */
#define E_INPCNT    (-11)       /* no usable inputs found */
#define E_VECSIZE   (-12)       /* invalid data point/vector size */
#define E_CONJFN    (-13)       /* unknown conjunction function */
#define E_DIFFCNT   (-14)       /* numbers of clusters differ */

#define SEC_SINCE(t)  ((double)(clock()-(t)) /(double)CLOCKS_PER_SEC)

/*--------------------------------------------------------------------*/
#define int         1           /* to check definitions */
#define long        2           /* for certain types */
#define ptrdiff_t   3

#if DIMID!=ATTID
#error "DIMID and ATTID must be the same data type"
#endif

#undef int                      /* remove preprocessor definitions */
#undef long                     /* needed for the type checking */
#undef ptrdiff_t

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- a partition matrix --- */
  DIMID  rowcnt;                /* number of rows (clusters) */
  DIMID  colcnt;                /* (current) number of columns */
  DIMID  colvsz;                /* size of the column vectors */
  double *wgts;                 /* columns weights */
  double *rows[1];              /* vector of rows */
} PARTMAT;                      /* (partition matrix) */

typedef struct {                /* --- mode information --- */
  int  code;                    /* code of mode */
  char *name;                   /* name of mode */
  char *desc;                   /* description */
} MODEINFO;                     /* (mode information) */

typedef double CONJFN (double a, double b);

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
static const char *errmsgs[] = {   /* error messages */
  /* E_NONE      0 */  "no error",
  /* E_NOMEM    -1 */  "not enough memory",
  /* E_FOPEN    -2 */  "cannot open file %s",
  /* E_FREAD    -3 */  "read error on file %s",
  /* E_FWRITE   -4 */  "write error on file %s",
  /* E_STDIN    -5 */  "double assignment of standard input",
  /* E_OPTION   -6 */  "unknown option -%c",
  /* E_OPTARG   -7 */  "missing option argument",
  /* E_ARGCNT   -8 */  "wrong number of arguments",
  /* E_PARSE    -9 */  "parse error(s) on file %s",
  /* E_ATTCNT  -10 */  "no (usable) attributes (need at least 1)",
  /* E_INPCNT  -11 */  "no (usable) inputs (need at least 1)",
  /* E_VECSIZE -12 */  "invalid vector size %"DIMID_FMT,
  /* E_CONJFN  -13 */  "unknown conjunction %s",
  /* E_DIFFCNT -14 */  "number of clusters differs "
                         "(%"DIMID_FMT" vs. %"DIMID_FMT")",
  /*           -15 */  "unknown error",
};

static const MODEINFO conjtab[] = { /* table of conjunction functions */
  { PM_MIN,   "min",   "minimum"             },
  { PM_MINNP, "minnp", "nilpotent minimum"   },
  { PM_MINNP, "npmin", NULL                  },
  { PM_PROD,  "prod",  "product"             },
  { PM_LUKA,  "luka",  "Lukasiewicz t-norm"  },
  { -1,       NULL,    NULL   /* sentinel */ },
};

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
static CCHAR   *prgname;        /* program name for error messages */
static SCANNER *scan   = NULL;  /* scanner for cluster set desc. */
static TABREAD *tread  = NULL;  /* table reader */
static ATTSET  *attset = NULL;  /* attribute set */
static ATTMAP  *attmap = NULL;  /* attribute map */
static CLSET   *clset1 = NULL;  /* first  cluster set */
static CLSET   *clset2 = NULL;  /* second cluster set */
static PARTMAT *pmat1  = NULL;  /* first  partition matrix */
static PARTMAT *pmat2  = NULL;  /* second partition matrix */

/*----------------------------------------------------------------------
  Conjunction Functions (t-norms)
----------------------------------------------------------------------*/

static double tmin (double a, double b)
{ return (a <= b) ? a : b; }

static double minnp (double a, double b)
{ return (a+b < 1) ? 0 : ((a <= b) ? a : b); }

static double prod (double a, double b)
{ return a * b; }

static double luka (double a, double b)
{ double t = a +b -1; return (t >= 0) ? t : 0; }

static CONJFN *cftab[] = { tmin, minnp, prod, luka, };

/*----------------------------------------------------------------------
  Partition Matrix Functions
----------------------------------------------------------------------*/

static PARTMAT* pm_create (DIMID clscnt)
{                               /* --- create a partition matrix */
  PARTMAT *pm;                  /* created partition matrix */

  pm = (PARTMAT*)malloc(sizeof(PARTMAT)+(size_t)clscnt*sizeof(double*));
  if (!pm) return NULL;         /* create the base structure */
  pm->rowcnt = clscnt;          /* store the number of rows */
  pm->colcnt = pm->colvsz = 0;  /* clear the column counter */
  pm->wgts   = NULL;            /* and the column weight array */
  while (--clscnt >= 0)         /* clear the column vectors */
    pm->rows[clscnt] = NULL;    /* (will be allocated when needed) */
  return pm;                    /* return the created matrix */
}  /* pm_create() */

/*--------------------------------------------------------------------*/

static int pm_addcol (PARTMAT *pm, CLSET *clset, double wgt)
{                               /* --- add current assig. to matrix */
  DIMID  i, vsz;                /* loop variable, new vector size */
  double *row;                  /* buffer for (re)allocation */

  assert(pm->rowcnt == cls_clscnt(clset));
  vsz = pm->colvsz;             /* get the size of the column vectors */
  if (pm->colcnt >= vsz) {      /* if the column vectors are full */
    vsz += (vsz > BLKSIZE) ? vsz >> 1 : BLKSIZE;
    row = (double*)realloc(pm->wgts, (size_t)vsz *sizeof(double));
    if (!row) return -1;        /* enlarge the weight vector */
    pm->wgts = row;             /* on sucess set the new vector */
    for (i = pm->rowcnt; --i >= 0; ) {
      row = (double*)realloc(pm->rows[i], (size_t)vsz *sizeof(double));
      if (!row) return -1;      /* enlarge a column vector and */
      pm->rows[i] = row;        /* on success set the new vector */
    } pm->colvsz = vsz;         /* store the new vector size */
  }                             /* (only after full success) */
  for (i = pm->rowcnt; --i >= 0; )
    pm->rows[i][pm->colcnt] = cls_msdeg(clset, i);
  pm->wgts[pm->colcnt++] = wgt; /* store the membership degrees */
  return 0;                     /* and the data point weight */
}  /* pm_addcol() */

/*--------------------------------------------------------------------*/

static int pm_cmp (PARTMAT *pm1, PARTMAT *pm2, int cfc1, int cfc2,
                   double *eval)
{                               /* --- compare partition matrices */
  DIMID  i, k, n;               /* loop variables, dummy */
  double *r1, *r2;              /* rows of the partition matrices */
  MATRIX *cmp;                  /* cluster comparison matrix */
  double n00, n01, n10, n11;    /* contingency table entries */
  double c1, c2;                /* coincidence matrix entries */
  double t1, t2;                /* temporary buffers */
  double wgt;                   /* total data point weight */
  CONJFN *conj1, *conj2;        /* conjunction functions (t-norms) */

  assert((pm1->rowcnt == pm2->rowcnt)
      && (pm1->colcnt == pm2->colcnt));
  cmp = mat_create(pm1->rowcnt, pm1->rowcnt);
  if (!cmp) return -1;          /* create a comparison matrix */

  /* --- compute total column weight --- */
  for (wgt = 0, n = pm1->colcnt; --n >= 0; ) {
    assert(pm1->wgts[n] == pm2->wgts[n]);
    wgt += pm1->wgts[n];        /* sum the column weights */
  }                             /* (total data point weights) */

  /* --- sum of squared differences --- */
  for (i = pm1->rowcnt; --i >= 0; ) {
    r1 = pm1->rows[i];          /* traverse rows of first matrix */
    for (k = pm2->rowcnt; --k >= 0; ) {
      r2 = pm2->rows[k];        /* traverse rows of second matrix */
      t1 = 0;                   /* init. sum of squared differences */
      for (n = pm1->colcnt; --n >= 0; ) {
        t2 = r1[n] -r2[n]; t1 += t2*t2 *pm1->wgts[n]; }
      mat_set(cmp, i, k, t1);   /* compute the sum of squared diffs. */
    }                           /* and store them in the matrix, */
  }                             /* then find minimum cost matching */
  eval[PM_DIFF] = mat_match(cmp, NULL) / wgt / (double)pm1->rowcnt;
  if (!(eval[PM_DIFF] > 0)) eval[PM_DIFF] = 0;
  if ((cfc1 < 0) || (cfc2 < 0)){/* if only mean squared difference, */
    mat_delete(cmp); return 0; }/* abort the function */

  /* --- cross-classification accuracy --- */
  conj1 = cftab[cfc1];          /* get the conjunction functions */
  conj2 = cftab[cfc2];          /* combining the membership degrees */
  for (i = pm1->rowcnt; --i >= 0; ) {
    r1 = pm1->rows[i];          /* traverse rows of first matrix */
    for (k = pm1->rowcnt; --k >= 0; ) {
      r2 = pm2->rows[k];        /* traverse rows of second matrix */
      n11 = n00 = 0;            /* init. contingency table elements */
      for (n = pm1->colcnt; --n >= 0; ) {
        n11 += conj1(  r1[n],   r2[n]) *pm1->wgts[n];
        n00 += conj1(1-r1[n], 1-r2[n]) *pm1->wgts[n];
      }                         /* sum conjunction of assignments */
      mat_set(cmp, i, k, -n11 -n00);
    }                           /* compute cross-class. accuracies */
  }                             /* and store them in the matrix */
  eval[PM_ACC] = -mat_match(cmp, NULL) / wgt / (double)pm1->rowcnt;
  if (!(eval[PM_ACC] > 0)) eval[PM_ACC] = 0;

  /* --- F1 measure --- */
  for (i = pm1->rowcnt; --i >= 0; ) {
    r1 = pm1->rows[i];          /* traverse rows of first matrix */
    for (k = pm1->rowcnt; --k >= 0; ) {
      r2 = pm2->rows[k];        /* traverse rows of second matrix */
      n11 = n01 = n10 = 0;      /* init. contingency table elements */
      for (n = pm1->colcnt; --n >= 0; ) {
        n11 += conj1(  r1[n],   r2[n]) *pm1->wgts[n];
        n01 += conj1(1-r1[n],   r2[n]) *pm1->wgts[n];
        n10 += conj1(  r1[n], 1-r2[n]) *pm1->wgts[n];
      }                         /* compute contingency table entries */
      mat_set(cmp, i, k, -2*n11 /(n10 +n01 +2*n11));
    }                           /* compute the F1 measures */
  }                             /* and store them in the matrix */
  eval[PM_F1] = -mat_match(cmp, NULL) / (double)pm1->rowcnt;
  if (!(eval[PM_F1] > 0)) eval[PM_F1] = 0;
  mat_delete(cmp);              /* evaluate all permutations */

  /* --- coincidence matrix measures --- */
  n11 = n10 = n01 = n00 = 0;    /* clear the contingency table */
  for (i = pm1->colcnt; --i > 0; ) {
    for (k = i; --k >= 0; ) {   /* traverse all pairs of tuples */
      c1 = c2 = 0;              /* init. coincidence matrix elements */
      for (n = pm1->rowcnt; --n >= 0; ) {
        c1 += conj1(pm1->rows[n][i], pm1->rows[n][k]) *pm1->wgts[n];
        c2 += conj1(pm2->rows[n][i], pm2->rows[n][k]) *pm2->wgts[n];
      }                         /* compute coincidence matrix entries */
      n11 += conj2(  c1,   c2); /* and from it the four entries */
      n10 += conj2(  c1, 1-c2); /* of the contingency table */
      n01 += conj2(1-c1,   c2); /* (assignment of data points */
      n00 += conj2(1-c1, 1-c2); /* to same or different clusters) */
    }
  }
  t1 = (n11 +n10) *(n11 +n01);  /* compute some useful values that */
  t2 = (n00 +n10) *(n00 +n01);  /* appear in more than one measure */
  c1 = (n11 +n10 +n01 +n00);    /* and compute the measures */
  eval[PM_RAND]    = (n11 +n00) / c1;
  eval[PM_JACCARD] = n11 / (n11 +n10 +n01);
  eval[PM_FOLKES]  = n11 / sqrt(t1);
  eval[PM_HUBERT]  = (c1*n11 -t1) /sqrt(t1 *t2);
  return 0;                     /* return 'ok' */
}  /* pm_cmp() */

/*--------------------------------------------------------------------*/
#ifndef NDEBUG

static void pm_delete (PARTMAT *pm)
{                               /* --- add current assig. to matrix */
  DIMID i;                      /* loop variable */

  if (pm->wgts) free(pm->wgts); /* delete the column weights */
  for (i = pm->rowcnt; --i >= 0; )
    free(pm->rows[i]);          /* delete the matrix rows */
  free(pm);                     /* and the base structure */
}  /* pm_delete() */

#endif
/*----------------------------------------------------------------------
  Main Functions
----------------------------------------------------------------------*/

#ifndef NDEBUG                  /* if debug version */
  #undef  CLEANUP               /* clean up memory and close files */
  #define CLEANUP \
  if (pmat1)  pm_delete(pmat1);     \
  if (pmat2)  pm_delete(pmat2);     \
  if (clset1) cls_delete(clset1);   \
  if (clset2) cls_delete(clset2);   \
  if (attmap) am_delete(attmap, 0); \
  if (attset) as_delete(attset);    \
  if (tread)  trd_delete(tread, 1); \
  if (scan)   scn_delete(scan,  1);
#endif

GENERROR(error, exit)           /* generic error reporting function */

/*--------------------------------------------------------------------*/

static void help (void)
{                               /* --- print help on init. modes */
  int i;                        /* loop variable */

  fprintf(stderr, "\n");        /* terminate startup message */
  printf("list of conjunction functions (options -m# and -c#):\n");
  printf("  name        conjunction function\n");
  for (i = 0; conjtab[i].name; i++)
    if (conjtab[i].desc)        /* list of conjunction functions */
      printf("  %-10s  %s\n", conjtab[i].name, conjtab[i].desc);
  exit(0);                      /* abort the program */
}  /* help() */

/*--------------------------------------------------------------------*/

static int code (const MODEINFO *tab, const char *name)
{                               /* --- get code of processing mode */
  for ( ; tab->name; tab++)     /* look up name in table */
    if (strcmp(tab->name, name) == 0)
      return tab->code;         /* return the code of the init. mode */
  return -1;                    /* or an error indicator */
}  /* code() */

/*--------------------------------------------------------------------*/

int main (int argc, char *argv[])
{                               /* --- main function */
  int     i, k = 0;             /* loop variables, counters */
  char    *s;                   /* to traverse options */
  CCHAR   **optarg = NULL;      /* option argument */
  CCHAR   *fn_cls1 = NULL;      /* name of first  cluster set file */
  CCHAR   *fn_cls2 = NULL;      /* name of second cluster set file */
  CCHAR   *fn_hdr  = NULL;      /* name of table header file */
  CCHAR   *fn_inp  = NULL;      /* name of table file */
  CCHAR   *recseps = NULL;      /* record  separators */
  CCHAR   *fldseps = NULL;      /* field   separators */
  CCHAR   *blanks  = NULL;      /* blank   characters */
  CCHAR   *comment = NULL;      /* comment characters */
  CCHAR   *conjfn1 = "prod";    /* conjunction functions for */
  CCHAR   *conjfn2 = "prod";    /* combining membership degrees */
  int     matinp   =  0;        /* flag for numerical matrix input */
  int     mode     = AS_ATT|AS_NOXATT|AS_NONULL;/* table read mode */
  int     donly    = 0;         /* flag for only mean squared diff. */
  ATTSET  *dummy   = NULL;      /* attribute set (dummy for reading) */
  int     cfc1, cfc2;           /* conjunction function codes */
  DIMID   clscnt;               /* number of clusters */
  double  eval[8];              /* result of comparison */
  double  *vec;                 /* to traverse the data points */
  ATTID   a, c;                 /* number of attributes */
  DIMID   m, o;                 /* number of inputs/dimensions */
  DIMID   n;                    /* number of data points/vectors */
  double  w, u;                 /* weight of data points/vectors */
  clock_t t;                    /* timer for measurements */

  prgname = argv[0];            /* get program name for error msgs. */

  /* --- print startup/usage message --- */
  if (argc > 1) {               /* if arguments are given */
    fprintf(stderr, "%s - %s\n", argv[0], DESCRIPTION);
    fprintf(stderr, VERSION); } /* print a startup message */
  else {                        /* if no argument is given */
    printf("usage: %s [options] clsfile1 clsfile2 "
                     "[-d|-h hdrfile] tabfile\n", argv[0]);
    printf("%s\n", DESCRIPTION);
    printf("%s\n", VERSION);
    printf("-x       compute only mean squared difference\n");
    printf("-m#      conjunction for membership degrees     "
                    "(default: product)\n");
    printf("-c#      conjunction for coincidence matrix     "
                    "(default: product)\n");
    printf("-r#      record  separators                     "
                    "(default: \"\\n\")\n");
    printf("-f#      field   separators                     "
                    "(default: \" \\t,\")\n");
    printf("-b#      blank   characters                     "
                    "(default: \" \\t\\r\")\n");
    printf("-C#      comment characters                     "
                    "(default: \"#\")\n");
    printf("-n       number of tuple occurrences in last field\n");
    printf("clsfile  file to read cluster set description from\n");
    printf("-d       use default header "
                    "(field names = field numbers)\n");
    printf("-h       read table header "
                    "(field names) from hdrfile\n");
    printf("hdrfile  file containing table header "
                    "(field names)\n");
    printf("tabfile  table file to read "
                    "(field names in first record)\n");
    return 0;                   /* print a usage message */
  }                             /* and abort the program */

  /* --- evaluate arguments --- */
  for (i = 1; i < argc; i++) {  /* traverse arguments */
    s = argv[i];                /* get option argument */
    if (optarg) { *optarg = s; optarg = NULL; continue; }
    if ((*s == '-') && *++s) {  /* -- if argument is an option */
      while (1) {               /* traverse characters */
        switch (*s++) {         /* evaluate option */
          case '!': help();                break;
          case 'r': optarg = &recseps;     break;
          case 'f': optarg = &fldseps;     break;
          case 'b': optarg = &blanks;      break;
          case 'C': optarg = &comment;     break;
          case 'x': donly  = 1;            break;
          case 'm': optarg = &conjfn1;     break;
          case 'c': optarg = &conjfn2;     break;
          case 'n': mode  |= AS_WEIGHT;    break;
          case 'd': mode  |= AS_DFLT;      break;
          case 'h': optarg = &fn_hdr;      break;
          default : error(E_OPTION, *--s); break;
        }                       /* set option variables */
        if (!*s) break;         /* if at end of string, abort loop */
        if (optarg) { *optarg = s; optarg = NULL; break; }
      } }                       /* get option argument */
    else {                      /* -- if argument is no option */
      switch (k++) {            /* evaluate non-option */
        case  0: fn_cls1 = s;     break;
        case  1: fn_cls2 = s;     break;
        case  2: fn_inp  = s;     break;
        default: error(E_ARGCNT); break;
      }                         /* note filenames */
    }
  }
  if (optarg) error(E_OPTARG);  /* check option argument and */
  if (k != 3) error(E_ARGCNT);  /* the number of arguments */
  if (fn_hdr && (strcmp(fn_hdr, "-") == 0))
    fn_hdr = "";                /* convert "-" to "" */
  i = ( fn_hdr  && !*fn_hdr) ? 1 : 0;
  if  (!fn_cls1 || !*fn_cls1) i++;
  if  (!fn_cls2 || !*fn_cls2) i++;
  if  (!fn_inp  || !*fn_inp ) i++;
  if (i > 1) error(E_STDIN);    /* stdin must not be used twice */
  if (donly) cfc1 = cfc2 = -1;  /* if only mean squared difference, */
  else {                        /* clear both confunctions */
    cfc1 = code(conjtab,conjfn1);
    if (cfc1 < 0) error(E_CONJFN, conjfn1);
    cfc2 = code(conjtab,conjfn2);
    if (cfc2 < 0) error(E_CONJFN, conjfn2);
  }                             /* otherwise get the conjunctions */
  fputc('\n', stderr);          /* terminate the startup message */

  /* --- parse first cluster set --- */
  scan = scn_create();          /* create a scanner */
  if (!scan) error(E_NOMEM);    /* for the cluster set */
  t = clock();                  /* start timer, open input file */
  if (scn_open(scan, NULL, fn_cls1) != 0)
    error(E_FOPEN, scn_name(scan));
  fprintf(stderr, "reading %s ... ", scn_name(scan));
  matinp = (scn_first(scan) == T_ID)
        && (strcmp(scn_value(scan), "dom") != 0);
  if (matinp) {                 /* if matrix version */
    clset1 = cls_parse(scan, 0, 1); /* parse cluster descriptions */
    if (!clset1 || !scn_eof(scan, 1))
      error(E_PARSE, scn_name(scan));
    m = cls_inpcnt(clset1);     /* get the number of inputs */
    fprintf(stderr, "[%"DIMID_FMT" input(s),", m); }
  else {                        /* if table version */
    attset = as_create("domains", att_delete);
    if (!attset) error(E_NOMEM);      /* create an attribute set */
    if (as_parse(attset, scan, AT_ALL, 1) != 0)
      error(E_PARSE, scn_name(scan)); /* parse domain descriptions */
    a = as_attcnt(attset);      /* parse domain descriptions */
    if (a < 1) error(E_ATTCNT); /* check for at least one attribute */
    attmap = am_create(attset, 0, 1.0);
    if (!attmap) error(E_NOMEM);/* create an attribute map */
    c = as_target(attset, NULL, -1);
    if (c < a)                  /* if there is a target attribute, */
      am_target(attmap, c);     /* set it in the attribute map */
    m = am_inpcnt(attmap);      /* get the number of dimensions */
    clset1 = cls_parse(scan, m, 1); /* parse cluster descriptions */
    if (!clset1 || !scn_eof(scan, 1))
      error(E_PARSE, scn_name(scan));
    fprintf(stderr, "[%"ATTID_FMT" attribute(s),", a);
  }
  if (m < 1) error(E_INPCNT);   /* check for at least one input */
  clscnt = cls_clscnt(clset1);  /* get the number of clusters */
  fprintf(stderr, " %"DIMID_FMT" cluster(s)]", clscnt);
  fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));
  cls_setup(clset1);            /* set up the cluster set */

  /* --- read second cluster set --- */
  t = clock();                  /* start timer, open input file */
  if (scn_open(scan, NULL, fn_cls2) != 0)
    error(E_FOPEN, scn_name(scan));
  fprintf(stderr, "reading %s ... ", scn_name(scan));
  if (matinp) {                 /* if matrix version */
    clset2 = cls_parse(scan, m, 1); /* parse cluster descriptions */
    if (!clset2 || !scn_eof(scan, 1))
      error(E_PARSE, scn_name(scan));
    fprintf(stderr, "[%"DIMID_FMT" input(s),", m); }
  else {                        /* if table version */
    dummy = as_create("domains", att_delete);
    if (!dummy) error(E_NOMEM);       /* create an attribute set */
    if (as_parse(dummy, scan, AT_ALL, 1) != 0)
      error(E_PARSE, scn_name(scan)); /* parse domain descriptions */
    clset2 = cls_parse(scan, m, 1);   /* parse cluster descriptions */
    if (!clset2 || !scn_eof(scan, 1))
      error(E_PARSE, scn_name(scan));
    fprintf(stderr, "[%"ATTID_FMT" attribute(s),", as_attcnt(dummy));
    as_delete(dummy);           /* delete the attribute set */
  }
  c = cls_clscnt(clset2);       /* get the number of clusters */
  fprintf(stderr, " %"DIMID_FMT" cluster(s)]",   c);
  fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));
  scn_delete(scan, 1);          /* delete the scanner and */
  scan = NULL;                  /* clear the scanner variable */
  if (c != clscnt) error(E_DIFFCNT, clscnt, c);
  cls_setup(clset2);            /* set up the cluster set */

  /* --- create partition matrices --- */
  pmat1 = pm_create(clscnt);    /* create a partition matrix */
  pmat2 = pm_create(clscnt);    /* for each of the cluster sets */
  if (!pmat1 && !pmat2) error(E_NOMEM);

  if (matinp) {                 /* if matrix version */
    /* --- process data points/vectors --- */
    tread = trd_create();       /* create a table reader and */
    if (!tread) error(E_NOMEM); /* set the separator characters */
    trd_allchs(tread, recseps, fldseps, blanks, "", comment);
    t = clock();                /* start timer, open input file */
    if (trd_open(tread, NULL, fn_inp) != 0)
      error(E_FOPEN, trd_name(tread));
    fprintf(stderr, "reading %s ... ", trd_name(tread));
    o = -1;                     /* read the first data point */
    k = vec_readx(&vec, &o, tread);   /* from the input file */
    if (k) error(k, TRD_INFO(tread));
    if (o < m) { free(vec); error(E_VECSIZE, o); }
    for (n = 0; k == 0; n++) {  /* data point/vector read loop */
      cls_exec(clset1,vec,NULL);/* execute the first cluster set */
      if (pm_addcol(pmat1, clset1, 1.0) < 0) {
        free(vec); error(E_NOMEM); }
      cls_exec(clset2,vec,NULL);/* execute the second cluster set */
      if (pm_addcol(pmat2, clset2, 1.0) < 0) {
        free(vec); error(E_NOMEM); }
      k = vec_read(vec, o, tread);
    }                           /* read the next data point/vector */
    free(vec);                  /* delete the vector buffer */
    if (k < 0) error(k, TRD_INFO(tread));
    trd_delete(tread, 1);       /* close the input file and */
    tread = NULL;               /* delete the table reader */
    fprintf(stderr, "[%"DIMID_FMT" vector(s)]", n);
    fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));
  }

  else {                        /* if table version */
    /* --- read table header --- */
    tread = trd_create();       /* create a table reader and */
    if (!tread) error(E_NOMEM); /* set the separator characters */
    trd_allchs(tread, recseps, fldseps, blanks, "", comment);
    if (fn_hdr) {               /* if a header file is given */
      t = clock();              /* start timer, open input file */
      if (trd_open(tread, NULL, fn_hdr) != 0)
        error(E_FOPEN, trd_name(tread));
      fprintf(stderr, "reading %s ... ", trd_name(tread));
      k = as_read(attset, tread, (mode & ~AS_DFLT) | AS_ATT);
      if (k < 0) error(-k, as_errmsg(attset, NULL, 0));
      trd_close(tread);         /* read table header, close file */
      fprintf(stderr, "[%"ATTID_FMT" attribute(s)]", a);
      fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));
      mode &= ~(AS_ATT|AS_DFLT);/* print a success message and */
    }                           /* remove the attribute flag */

    /* --- process table body --- */
    t = clock();                /* start timer, open input file */
    if (trd_open(tread, NULL, fn_inp) != 0)
      error(E_FOPEN, trd_name(tread));
    fprintf(stderr, "reading %s ... ", trd_name(tread));
    k = as_read(attset, tread, mode);  /* read table header */
    if (k < 0) error(-k, as_errmsg(attset, NULL, 0));
    i = mode; mode = (mode & ~(AS_DFLT|AS_ATT)) | AS_INST;
    if (!(i & AS_DFLT))         /* if not done yet, read first tuple */
      k = as_read(attset, tread, mode);
    vec = (double*)malloc((size_t)m *sizeof(double));
    if (!vec) error(E_NOMEM);   /* create a data point/vector buffer */
    for (w = 0, n = 0; k == 0; n++) {
      w += u = as_getwgt(attset);   /* sum the tuple weight */
      am_exec(attmap, AM_INPUTS, vec); /* convert to vector */
      cls_exec(clset1, vec, NULL); /* execute first cluster set */
      if (pm_addcol(pmat1, clset1, u) < 0) {
        free(vec); error(E_NOMEM); }
      cls_exec(clset2, vec, NULL); /* execute second cluster set */
      if (pm_addcol(pmat2, clset2, u) < 0) {
        free(vec); error(E_NOMEM); }
      k = as_read(attset, tread, mode);
    }                           /* try to read the next record */
    free(vec);                  /* delete the vector buffer */
    if (k < 0) error(-k, as_errmsg(attset, NULL, 0));
    trd_delete(tread, 1);       /* delete the table reader */
    tread = NULL;               /* and clear the variable */
    fprintf(stderr, "[%"ATTID_FMT" attribute(s),", a);
    fprintf(stderr, " %"DIMID_FMT, n);
    if (w != (double)n) fprintf(stderr, "/%g", w);
    fprintf(stderr, " tuple(s)] done [%.2fs].\n", SEC_SINCE(t));
  }                             /* if (matinp) .. else .. */

  /* --- compute and print evaluations --- */
  for (i = 8; --i >= 0; ) eval[i] = 0;
  if (pm_cmp(pmat1, pmat2, cfc1, cfc2, eval) < 0)
    error(E_NOMEM);             /* compare the partition matrices */
  printf("mean squared difference      : %18.16g\n", eval[PM_DIFF]);
  if (!donly) {                 /* check for other measures */
  printf("cross-classification accuracy: %18.16g\n", eval[PM_ACC]);
  printf("F1 measure                   : %18.16g\n", eval[PM_F1]);
  printf("Rand statistic               : %18.16g\n", eval[PM_RAND]);
  printf("Jaccard coefficient          : %18.16g\n", eval[PM_JACCARD]);
  printf("Folkes-Mallows index         : %18.16g\n", eval[PM_FOLKES]);
  printf("Hubert index                 : %18.16g\n", eval[PM_HUBERT]); }

  /* --- clean up --- */
  CLEANUP;                      /* clean up memory and close files */
  SHOWMEM;                      /* show (final) memory usage */
  return 0;                     /* return 'ok' */
}  /* main() */
