/*----------------------------------------------------------------------
  File    : cle.c
  Contents: probabilistic and fuzzy cluster evaluation
  Author  : Christian Borgelt
  History : 2003.05.17 file created
            2003.05.18 setting of noise parameter (option -y) added
            2003.06.09 coverage of data added
            2003.08.16 slight changes in error message output
            2004.02.25 source files mcle.c and cle.c combined
            2006.01.28 conditional compilation simplified
            2007.02.14 adapted to changed module tabread
            2007.03.16 table and matrix version combined
            2013.08.16 adapted to definitions ATTID, VALID, DIMID etc.
            2013.08.29 adapted to new function as_target()
            2014.10.24 changed from LGPL license to MIT license
            2018.08.18 extended functions (attset/table) removed
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <assert.h>
#ifndef MAT_RDWR
#define MAT_RDWR
#endif
#include "matrix.h"
#ifndef AS_READ
#define AS_READ
#endif
#ifndef AS_WRITE
#define AS_WRITE
#endif
#ifndef AS_PARSE
#define AS_PARSE
#endif
#include "attset.h"
#include "attmap.h"
#ifndef CLS_PARSE
#define CLS_PARSE
#endif
#include "cluster.h"
#include "error.h"
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define PRGNAME     "cle"
#define DESCRIPTION "probabilistic and fuzzy cluster evaluation"
#define VERSION     "version 3.3 (2022.04.13)         " \
                    "(c) 2003-2022   Christian Borgelt"

/* --- error codes --- */
/* error codes 0 to -5 defined in attset.h */
#define E_OPTION     (-6)       /* unknown option */
#define E_OPTARG     (-7)       /* missing option argument */
#define E_ARGCNT     (-8)       /* wrong number of arguments */
#define E_PARSE      (-9)       /* parse errors on input file */
#define E_ATTCNT    (-10)       /* no usable attributes found */
#define E_INPCNT    (-11)       /* no usable inputs found */
#define E_VECSIZE   (-12)       /* invalid data point/vector size */

#define SEC_SINCE(t)  ((double)(clock()-(t)) /(double)CLOCKS_PER_SEC)

/*--------------------------------------------------------------------*/
#define int         1           /* to check definitions */
#define long        2           /* for certain types */
#define ptrdiff_t   3

#if DIMID!=ATTID
#error "DIMID and ATTID must be the same data type"
#endif

#undef int                      /* remove preprocessor definitions */
#undef long                     /* needed for the type checking */
#undef ptrdiff_t

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
static const char *errmsgs[] = {   /* error messages */
  /* E_NONE      0 */  "no error",
  /* E_NOMEM    -1 */  "not enough memory",
  /* E_FOPEN    -2 */  "cannot open file %s",
  /* E_FREAD    -3 */  "read error on file %s",
  /* E_FWRITE   -4 */  "write error on file %s",
  /* E_STDIN    -5 */  "double assignment of standard input",
  /* E_OPTION   -6 */  "unknown option -%c",
  /* E_OPTARG   -7 */  "missing option argument",
  /* E_ARGCNT   -8 */  "wrong number of arguments",
  /* E_PARSE    -9 */  "parse error(s) on file %s",
  /* E_ATTCNT  -10 */  "no (usable) attributes (need at least 1)",
  /* E_INPCNT  -11 */  "no (usable) inputs (need at least 1)",
  /* E_VECSIZE -12 */  "invalid vector size %"DIMID_FMT,
  /*           -13 */  "unknown error",
};

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
static CCHAR   *prgname;        /* program name for error messages */
static SCANNER *scan   = NULL;  /* scanner for cluster set desc. */
static TABREAD *tread  = NULL;  /* table reader */
static ATTSET  *attset = NULL;  /* attribute set */
static ATTMAP  *attmap = NULL;  /* attribute map */
static CLSET   *clset  = NULL;  /* cluster set */

/*----------------------------------------------------------------------
  Main Functions
----------------------------------------------------------------------*/

#ifndef NDEBUG                  /* if debug version */
  #undef  CLEANUP               /* clean up memory and close files */
  #define CLEANUP \
  if (clset)  cls_delete(clset);    \
  if (attmap) am_delete(attmap, 0); \
  if (attset) as_delete(attset);    \
  if (tread)  trd_delete(tread, 1); \
  if (scan)   scn_delete(scan,  1);
#endif

GENERROR(error, exit)           /* generic error reporting function */

/*--------------------------------------------------------------------*/

int main (int argc, char *argv[])
{                               /* --- main function */
  int     i, k = 0;             /* loop variables, counters */
  char    *s;                   /* to traverse options */
  CCHAR   **optarg = NULL;      /* option argument */
  CCHAR   *fn_cls  = NULL;      /* name of cluster set file */
  CCHAR   *fn_hdr  = NULL;      /* name of table header file */
  CCHAR   *fn_inp  = NULL;      /* name of input table file */
  CCHAR   *recseps = NULL;      /* record  separators */
  CCHAR   *fldseps = NULL;      /* field   separators */
  CCHAR   *blanks  = NULL;      /* blank   characters */
  CCHAR   *comment = NULL;      /* comment characters */
  int     matinp   =  0;        /* flag for numerical matrix input */
  double  radius   =  1;        /* radius for (avg.) part. density */
  int     mode     = AS_ATT|AS_NOXATT|AS_NONULL;/* table read mode */
  double  *vec;                 /* to traverse the data points */
  DIMID   clscnt;               /* number of clusters */
  DIMID   a, c;                 /* number of attributes */
  DIMID   m, o;                 /* number of inputs/dimensions */
  DIMID   n;                    /* number of data points/vectors */
  double  w, u;                 /* weight of data points/vectors */
  clock_t t;                    /* timer for measurements */

  prgname = argv[0];            /* get program name for error msgs. */

  /* --- print startup/usage message --- */
  if (argc > 1) {               /* if arguments are given */
    fprintf(stderr, "%s - %s\n", argv[0], DESCRIPTION);
    fprintf(stderr, VERSION); } /* print a startup message */
  else {                        /* if no argument is given */
    printf("usage: %s [options] clsfile [-d|-h hdrfile] "
                     "tabfile\n", argv[0]);
    printf("%s\n", DESCRIPTION);
    printf("%s\n", VERSION);
    printf("-p#      radius for partition density           "
                    "(default: %g)\n", radius);
    printf("-C#      comment characters                     "
                    "(default: \"#\")\n");
    printf("-r#      record  separators                     "
                    "(default: \"\\n\")\n");
    printf("-f#      field   separators                     "
                    "(default: \" \\t,\")\n");
    printf("-b#      blank   characters                     "
                    "(default: \" \\t\\r\")\n");
    printf("-C#      comment characters                     "
                    "(default: \"#\")\n");
    printf("-n       number of tuple occurrences in last field\n");
    printf("clsfile  file to read cluster set description from\n");
    printf("-d       use default header "
                    "(field names = field numbers)\n");
    printf("-h       read table header "
                    "(field names) from hdrfile\n");
    printf("hdrfile  file containing table header "
                    "(field names)\n");
    printf("tabfile  table file to read "
                    "(field names in first record)\n");
    return 0;                   /* print a usage message */
  }                             /* and abort the program */

  /* --- evaluate arguments --- */
  for (i = 1; i < argc; i++) {  /* traverse arguments */
    s = argv[i];                /* get option argument */
    if (optarg) { *optarg = s; optarg = NULL; continue; }
    if ((*s == '-') && *++s) {  /* -- if argument is an option */
      while (1) {               /* traverse characters */
        switch (*s++) {         /* evaluate option */
          case 'p': radius = strtod(s, &s); break;
          case 'r': optarg = &recseps;      break;
          case 'f': optarg = &fldseps;      break;
          case 'b': optarg = &blanks;       break;
          case 'C': optarg = &comment;      break;
          case 'n': mode  |= AS_WEIGHT;     break;
          case 'd': mode  |= AS_DFLT;       break;
          case 'h': optarg = &fn_hdr;       break;
          default : error(E_OPTION, *--s);  break;
        }                       /* set option variables */
        if (!*s) break;         /* if at end of string, abort loop */
        if (optarg) { *optarg = s; optarg = NULL; break; }
      } }                       /* get option argument */
    else {                      /* -- if argument is no option */
      switch (k++) {            /* evaluate non-option */
        case  0: fn_cls = s;      break;
        case  1: fn_inp = s;      break;
        default: error(E_ARGCNT); break;
      }                         /* note filenames */
    }
  }
  if (optarg) error(E_OPTARG);  /* check option argument and */
  if (k != 2) error(E_ARGCNT);  /* the number of arguments */
  if (fn_hdr && (strcmp(fn_hdr, "-") == 0))
    fn_hdr = "";                /* convert "-" to "" */
  i = ( fn_hdr && !*fn_hdr) ? 1 : 0;
  if  (!fn_inp || !*fn_inp) i++;
  if (i > 1) error(E_STDIN);    /* stdin must not be used twice */
  fputc('\n', stderr);          /* terminate the startup message */

  /* --- parse cluster set --- */
  scan = scn_create();          /* create a scanner */
  if (!scan) error(E_NOMEM);    /* for the cluster set */
  t = clock();                  /* start timer, open input file */
  if (scn_open(scan, NULL, fn_cls) != 0)
    error(E_FOPEN, scn_name(scan));
  fprintf(stderr, "reading %s ... ", scn_name(scan));
  matinp = (scn_first(scan) == T_ID)
        && (strcmp(scn_value(scan), "dom") != 0);
  if (matinp) {                 /* if matrix version */
    clset = cls_parse(scan,0,1);/* parse cluster descriptions */
    if (!clset || !scn_eof(scan, 1))
      error(E_PARSE, scn_name(scan));
    m = cls_inpcnt(clset);      /* get the number of inputs */
    fprintf(stderr, "[%"DIMID_FMT" input(s),", m); }
  else {                        /* if table version */
    attset = as_create("domains", att_delete);
    if (!attset) error(E_NOMEM);      /* create an attribute set */
    if ((as_parse(attset, scan, AT_ALL, 1) != 0))
      error(E_PARSE, scn_name(scan)); /* parse domain descriptions */
    a = as_attcnt(attset);      /* parse domain descriptions */
    if (a < 1) error(E_ATTCNT); /* check for at least one attribute */
    attmap = am_create(attset, 0, 1.0);
    if (!attmap) error(E_NOMEM);/* create an attribute map */
    a = as_attcnt(attset);      /* get the number of attributes */
    c = as_target(attset, NULL, -1);
    if (c < a)                  /* if there is a target attribute, */
      am_target(attmap, c);     /* set it in the attribute map */
    m = am_inpcnt(attmap);      /* get the number of inputs */
    clset = cls_parse(scan,m,1);/* parse cluster descriptions */
    if (!clset || !scn_eof(scan, 1))
      error(E_PARSE, scn_name(scan));
    fprintf(stderr, "[%"ATTID_FMT" attribute(s),", a);
  }
  if (m < 1) error(E_INPCNT);   /* check for at least one input */
  clscnt = cls_clscnt(clset);   /* get the number of clusters */
  fprintf(stderr, " %"DIMID_FMT" cluster(s)]", clscnt);
  fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));
  scn_delete(scan, 1);          /* delete the scanner and */
  scan = NULL;                  /* clear the scanner variable */
  cls_setup(clset);             /* set up the cluster set */
  for (i = CLS_VMCNT; --i >= 0; )
    cls_evcfg(clset,i,radius);  /* configure the evaluation */

  if (matinp) {                 /* if matrix version */
    /* --- process data points/vectors --- */
    tread = trd_create();       /* create a table reader and */
    if (!tread) error(E_NOMEM); /* set the separator characters */
    trd_allchs(tread, recseps, fldseps, blanks, "", comment);
    t = clock();                /* start timer, open input file */
    if (trd_open(tread, NULL, fn_inp) != 0)
      error(E_FOPEN, trd_name(tread));
    fprintf(stderr, "reading %s ... ", trd_name(tread));
    o = -1;                     /* read the first data point */
    k = vec_readx(&vec, &o, tread);   /* from the input file */
    if (k) error(k, TRD_INFO(tread)); /* from the input file */
    if (o < m) { free(vec); error(E_VECSIZE, o); }
    for (n = 0; k == 0; n++) {  /* data point/vector read loop */
      cls_evaggr(clset, vec, 1);/* aggregate the data point/vector */
      k = vec_read(vec, o, tread);
    }                           /* read the next data point/vector */
    free(vec);                  /* delete the vector buffer */
    if (k < 0) error(k, TRD_INFO(tread));
    trd_delete(tread, 1);       /* close the input file and */
    tread = NULL;               /* delete the table reader */
    fprintf(stderr, "[%"DIMID_FMT" vector(s)]", n);
    fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));
  }

  else {                        /* if table version */
    /* --- read table header --- */
    tread = trd_create();       /* create a table reader and */
    if (!tread) error(E_NOMEM); /* set the separator characters */
    trd_allchs(tread, recseps, fldseps, blanks, "", comment);
    if (fn_hdr) {               /* if a header file is given */
      t = clock();              /* start timer, open input file */
      if (trd_open(tread, NULL, fn_hdr) != 0)
        error(E_FOPEN, trd_name(tread));
      fprintf(stderr, "reading %s ... ", trd_name(tread));
      k = as_read(attset, tread, (mode & ~AS_DFLT) | AS_ATT);
      if (k < 0) error(-k, as_errmsg(attset, NULL, 0));
      trd_close(tread);         /* read table header, close file */
      fprintf(stderr, "[%"ATTID_FMT" attribute(s)]", a);
      fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));
      mode &= ~(AS_ATT|AS_DFLT);/* print a success message and */
    }                           /* remove the attribute flag */

    /* --- process table body --- */
    t = clock();                /* start timer, open input file */
    if (trd_open(tread, NULL, fn_inp) != 0)
      error(E_FOPEN, trd_name(tread));
    fprintf(stderr, "reading %s ... ", trd_name(tread));
    k = as_read(attset, tread, mode);  /* read table header */
    if (k < 0) error(-k, as_errmsg(attset, NULL, 0));
    i = mode; mode = (mode & ~(AS_DFLT|AS_ATT)) | AS_INST;
    if (!(i & AS_DFLT))         /* if not done yet, read first tuple */
      k = as_read(attset, tread, mode);
    vec = (double*)malloc((size_t)m *sizeof(double));
    if (!vec) error(E_NOMEM);   /* create a data point/vector buffer */
    for (w = 0, n = 0; k == 0; n++) {
      w += u = as_getwgt(attset);   /* sum the tuple weight */
      am_exec(attmap, AM_INPUTS, vec); /* convert to vector */
      cls_evaggr(clset, vec, u);/* and aggregate the vector */
      k = as_read(attset, tread, mode);
    }                           /* try to read the next record */
    free(vec);                  /* delete the vector buffer */
    if (k < 0) error(-k, as_errmsg(attset, NULL, 0));
    trd_delete(tread, 1);       /* delete the table reader */
    tread = NULL;               /* and clear the variable */
    fprintf(stderr, "[%"ATTID_FMT" attribute(s),", a);
    fprintf(stderr, " %"DIMID_FMT, n);
    if (w != (double)n) fprintf(stderr, "/%g", w);
    fprintf(stderr, " tuple(s)] done [%.2fs].\n", SEC_SINCE(t));
  }                             /* if (matinp) .. else .. */

  /* --- compute and print evaluations --- */
  printf("objective function          (min): %g\n",
         cls_eval(clset, CLS_OBJFN));
  printf("coverage                    (max): %g\n",
         cls_eval(clset, CLS_COVERAGE));
  printf("fuzzy Davies-Bouldin index  (min): %g\n",
         cls_eval(clset, CLS_FDBIDX));
  printf("partition coefficient       (max): %g\n",
         cls_eval(clset, CLS_PCOEFF));
  printf("partition entropy           (min): %g\n",
         cls_eval(clset, CLS_PENTROPY));
  printf("fuzzy hyper-volume          (min): %g\n",
         cls_eval(clset, CLS_FHYPVOL));
  printf("proportion exponent         (max): %g\n",
         cls_eval(clset, CLS_PRPEXP));
  printf("Xie-Beni index / separation (min): %g\n",
         cls_eval(clset, CLS_XBISEP));
  printf("partition density           (max): %g\n",
         cls_eval(clset, CLS_PARTDENS));
  printf("average partition density   (max): %g\n",
         cls_eval(clset, CLS_PDAVG));

  /* --- clean up --- */
  CLEANUP;                      /* clean up memory and close files */
  SHOWMEM;                      /* show (final) memory usage */
  return 0;                     /* return 'ok' */
}  /* main() */
